/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: GRAB_UNF.c $
   $Locker: ska $	$Name:  $	$State: Exp $

	Grab NLS initialization data from current country information
	and dump it in an Uniform NLS datafile Format.

	The created files look like:
		ccc-ppp.unf
	where:
		ccc: current country code
		ppp: current code page

*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <stdarg.h>

#include "../unf.h"

#if defined(__TURBOC__) && !defined(__BORLANDC__)
# if __TURBOC__ < 0x0300
# include "missing.c"
# endif
#endif

	/* Has been renamed to avoid name clashes */
#define	error unf__eerror


#ifndef lint
static char const rcsid[] = 
	"$Id: GRAB_UNF.c 1.2 2000/07/02 04:51:06 ska Exp ska $";
#endif

struct nlsExtCntryInfo eci;

/* characters of Yes/No prompt for DOS-65-23 */
UWORD yesnoBuf[2];
#define yesCharacter yesnoBuf[0]
#define noCharacter yesnoBuf[1]

/* upcased characters for ECS-ASCII > 0x7f for DOS-65-02 */
struct nlsCharTbl256 normCh;
#define upNCsize normCh.numEntries
#define upNormCh normCh.tbl

/* lowercased characters for ECS-ASCII > 0x7f for DOS-65-03 */
struct nlsCharTbl256 lonormCh;
#define loNCsize lonormCh.numEntries
#define loNormCh lonormCh.tbl

/* upcased characters for ECS-ASCII > 0x7f for file names for DOS-65-04 */
struct nlsCharTbl256 fileCh;
#define upFCsize fileCh.numEntries
#define upFileCh fileCh.tbl

/* collating sequence for ECS-ASCII 0..0xff for DOS-65-06 */
struct nlsCharTbl256 coll;
#define collSize coll.numEntries
#define collSeq coll.tbl

/* DBC support for DOS-65-07 */
struct nlsDBCS dbcs;
#define dbcSize dbcs.numEntries
#define dbcEndMarker dbcs.dbcsTbl[0]

/* in file names permittable characters for DOS-65-05 */
#define nlsFnBuf 256
struct {
	struct nlsFnamTerm  nlsFnX;
	BYTE dummy[nlsFnBuf];		/* room to extend nlsFnX.separators to */
} nlsFnY;
#define nlsFn nlsFnY.nlsFnX

struct UNF_IN_BUFFER iBuf;
#define get(a) unf_get(a, &iBuf)

void mvTable(void *tbl, char *s, unsigned len)
{	unsigned i;

	memset(tbl, 0, len + 2);
	if(iBuf.info.pointer == 0) return;

	i = *(WORD far*)iBuf.info.pointer;
	if(i > len)
		error("Too many characters in %s table", s);
	_fmemcpy((void far*)tbl, iBuf.info.pointer, i + 2);
}


int main(void)
{	
	unsigned i;
	unsigned c;

		/* retreive extended country information */
	if(get(1)) {
		fputs("Cannot proceed, sorry!\n", stderr);
		return 20;
	}
		/* The extended country inormation contains 10 "dummy" bytes
			at its end */
	if(iBuf.info.subInfo.size > sizeof(eci) - 3 + 10) {
		fprintf(stderr, "ExtCntryInfo buffer overflow (%u)\n", iBuf.info.subInfo.size);
		return 30;
	}
	memcpy(&eci, &iBuf, sizeof(eci));

	/* Try to identify the yes/no characters.
		If the uppercased character is the same yes/no character,
		it is preferred to the lowercase character */
	for(i = yesCharacter = noCharacter = 0; i < 256; i++) {
#if defined(__TURBOC__)
		_DX = i;
		_AX = 0x6523;
		geninterrupt(0x21);
		switch(_AX) {
			case 0:
				_DX = noCharacter = i;
				_AX = 0x6520;
				geninterrupt(0x21);
				_DL = c = _AL;		/* uppercased character */
				_AX = 0x6523;
				geninterrupt(0x21);
				if(_AX == 0)		/* is the same --> use it */
					noCharacter = c;
				if(yesCharacter) goto brkSe;
				break;
			case 1:
				_DX = yesCharacter = i;
				_AX = 0x6520;
				geninterrupt(0x21);
				_DL = c = _AL;		/* uppercased character */
				_AX = 0x6523;
				geninterrupt(0x21);
				if(_AX == 1)		/* is the same --> use it */
					yesCharacter = c;
				if(noCharacter) goto brkSe;
				break;
		}
#else
        union REGS r;
        struct SREGS sr;
        r.x.dx = i;
        r.x.ax = 0x6523;
        intdosx(&r, &r, &sr);
        switch(r.x.ax) {
            case 0:
                r.x.dx = noCharacter = i;
                r.x.ax = 0x6520;
                intdosx(&r, &r ,&sr);
                r.h.dl = c = r.h.al;    /* uppercased character */
                r.x.ax = 0x6523;
                intdosx(&r, &r, &sr);
                if(r.x.ax == 0)         /* is the same --> use it */
                    noCharacter = c;
                if(yesCharacter) goto brkSe;
                break;
            case 1:
                r.x.dx = yesCharacter = i;
                r.x.ax = 0x6520;
                intdosx(&r, &r, &sr);
                r.h.dl = c = r.h.al;    /* uppercased character */
                r.x.ax = 0x6523;
                intdosx(&r, &r, &sr);
                if(r.x.ax == 1)         /* is the same --> use it */
                    yesCharacter = c;
                if(noCharacter) goto brkSe;

                break;
        }
#endif
	}
	fputs("Cannot determine yes/no prompts\n", stderr);
brkSe:

	get(2);			/* upcase table */
	mvTable(&upNCsize, "upcase", 256);
	if(!upNCsize)
		error("Failed to retreive upcase table 2");

	get(3);			/* upcase table */
	mvTable(&loNCsize, "locase", 256);

	get(4);			/* upcase table for file names */
	mvTable(&upFCsize, "file name upcase", 256);
	if(!upFCsize)
		error("Failed to retreive filename upcase table 4");

	get(6);			/* collating sequence */
	mvTable(&collSize, "collating", 256);
	if(!collSize)
		error("Failed to retreive collating sequence table 6");

	get(5);			/* permittable characters in file names */
	if(iBuf.info.pointer == 0) {
		fputs("Cannot continue, sorry!\n", stderr);
		return 23;
	}
	_fmemcpy((void far*)&nlsFn, iBuf.info.pointer, sizeof(nlsFn) + nlsFnBuf);

	get(7);			/* DBC support */
	if(iBuf.info.pointer && *(WORD FAR*)iBuf.info.pointer) {
		mvTable(&dbcs, "DBCS Lead Byte Tables", sizeof(dbcs.dbcsTbl));
	}
	else {
		/* On SBCS environment, sizeof(nlsDBCS) = 0 and nlsDBCSrgn = 0 */
		_fmemcpy((void far *)&dbcs, iBuf.info.pointer, 4);
	}

/* structure filled => generate NLS info file */
	switch(unf_create(NULL, eci.countryCode, eci.codePage)) {
	case 1:
		error("Cannot create NLS information file");
	case 2:
		error("Out of memory");
	default:
		error("Unknown error while creating UNF file");
	case 0:			/* OK */
		break;
	}

	/* Dump preample */
	unf_oradix(10);		/* Radix used to created output */
	unf_oend();			/* Global section ends */


	unf_oExtInf(&eci);			/* Dump table 1 */
	unf_oCharTab(2, "normal character uppercase table", &upNCsize);
	if(loNCsize)
		unf_oCharTab(3, "normal character lowercase table", &loNCsize);
	unf_oCharTab(4, "filename character uppercase table", &upFCsize);
	unf_oFTerm(&nlsFn);
	unf_oCharTab(6, "collating sequence table", &collSize);
	unf_oDBCS(&dbcSize);		/* Dump table 7 */
#if defined(DBCS)
	if (sizeof(((struct nlsPackage *)0)->yeschar) == 1) {
		char	tmpyn[2];
		tmpyn[0] = (char)(yesnoBuf[0]);
		tmpyn[1] = (char)(yesnoBuf[1]);
		unf_oYesNo(tmpyn);	/* Dump table 0x23 */
	} else {
		unf_oYesNo((char *)yesnoBuf);	/* Dump table 0x23 */
	}
#else
	unf_oYesNo((char *)yesnoBuf);	/* Dump table 0x23 */
#endif
	unf_close();

	printf("NLS info file for %03u-%03u has been created successfully.\n"
	 , eci.countryCode, eci.codePage);
	return 0;
}
