/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: Load_unf.c $
   $Locker: ska $	$Name:  $	$State: Exp $

	Read in an UNF file and loads into into memory permanently.
	Furtherly, it can be activated by DOS-38.

	Useage:	LOAD_UNF [/a] [filename]
		-or-
			LOAD_UNF [/a] country_code codepage_ID

		The optional switch "/a" automatically activates the NLS pkg.

		"filename" refers to the UNF file to be loaded.
		country_code and codepage_ID specify the NLS pkg with its
		numerical encoding.
		If the filename is missing or is "-", standard input is read.

	Note:
		This is an intermediate program to bridge the time until the
		COUNTRY= statement is fully included into the kernel and the
		fileformat of the COUNTRY.SYS is fixed.

	Note #2:
		The UNF must pass the kernel check of the data, that means it must
		conform to the structure that can be handled by the kernel without
		any external code.

 */

#if defined(__TURBOC__)
#include <dos.h>
#endif
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../version.h"
#include "../unf.h"
#include <dos.h>
#define _Carry (_FLAGS & 1)

#if !defined(MK_FP)
#define MK_FP(s,o) (void far *)(((unsigned long)(unsigned short)(s) << 16) | (unsigned short)(o))
#endif
#if !defined(FP_SEG)
#define FP_SEG(p) (unsigned)((unsigned long)(void far *)(p) >> 16)
#endif
#if !defined(FP_OFF)
#define FP_OFF(p) (unsigned short)(p)
#endif

	/* Define for testing under non-FreeDOS kernels */
//#define NON_FD_TEST
#define DEBUG

/* Declaration for interrupt functions */
typedef void interrupt (*fintr)();
typedef void interrupt (far *tintr)(unsigned bp, unsigned di, unsigned si, unsigned ds, unsigned es, unsigned dx, unsigned cx, unsigned bx, unsigned ax, unsigned ip, unsigned cs, unsigned flag);

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: Load_unf.c 1.1 2000/07/02 04:51:06 ska Exp ska $";
#endif

int fdVers = -1;			/* FreeDOS kernel version of NLS implementation */
tintr orgMUX = 0;			/* original MUX handler */
struct nlsPackage *pkg;		/* The main entry point of NLS pkgs */
struct nlsInfoBlock far* nlsRoot = 0;	/* the central NLS info block */
int activate = 0;			/* activate NLS pkg after loading */

#ifdef NON_FD_TEST
struct nlsInfoBlock nls_local_root = {
	 (char FAR *)0		/* filename to COUNTRY.SYS */
	,437				/* system code page */
	/* Implementation flags */
	,0
#ifdef NLS_MODIFYABLE_DATA
		| NLS_CODE_MODIFYABLE_DATA
#endif
#ifdef NLS_REORDER_POINTERS
		| NLS_CODE_REORDER_POINTERS
#endif
	,&nlsPackageHardcoded	/* hardcoded first package */
	,&nlsPackageHardcoded	/* first item in chain */
};
#endif

#pragma warn -par
#pragma warn -ofp
#pragma warn -rvl
void interrupt muxHandler(unsigned bp, unsigned di, unsigned si, 
	unsigned ds, unsigned es, unsigned dx, unsigned cx, unsigned bx,
	unsigned ax, unsigned ip, unsigned cs, unsigned flags)
{	
	if(ax == 0x1400 && cx == NLS_FREEDOS_NLSFUNC_ID) {
		/* MUX-14 from FreeDOS kernel */
		fdVers = bx;
		if(bx == NLS_FREEDOS_NLSFUNC_VERSION) {
			/* The correct FreeDOS kernel and the correct API call */
			nlsRoot = MK_FP(ds, si);
			flags |= 1;			/* Set carry flag */
			ax = 0x1401;	/* Return: "Not installed, not OK to install */
			return;
		}
	}
	_chain_intr((fintr)orgMUX);
}
#pragma warn +par
#pragma warn +ofp
#pragma warn +rvl


static int fetchNlsRoot(void)
{	if(nlsRoot)		/* already fecthed */
		return 1;

	orgMUX = (tintr)_dos_getvect(0x2f);
	_dos_setvect(0x2f, (fintr)muxHandler);
	_DX = 0xffff;
	_BX = 0xfffe;		/* invalid country code */
	_AX = 0x38ff;		/* Set Country code in BX */
	geninterrupt(0x21);
	_dos_setvect(0x2f, orgMUX);

#ifdef	NON_FD_TEST
	nlsRoot = &nls_local_root;
#else
	/* Now, we directly manipulate the NLS pkg chain */
	if(!nlsRoot) {
		if(fdVers != -1)
			fprintf(stderr, "This utility cannot run on this FreeDOS kernel.\n"
				"The kernel requested NLS structure version #%u,\n"
				"this tool only knows #%u.\n", fdVers
				 , NLS_FREEDOS_NLSFUNC_VERSION);
		else
			fputs("Sorry, this utility requires a FreeDOS kernel with NLS included.\n", stderr);
		return 0;
	}
#endif

	return 1;
}

/*	In the future this tool requires to know the address of the
	central nlsInfoBlock of the kernel.
	Then, it requires to know if the NLS pkg is already loaded.

	The method to do so is such:

	1) We hook MUX-14 (NLSFUNC).
	2) We initiate DOS-38 with an invalid country code.

	===

	_If_ LOAD_UNF was invoked ontop of a FreeDOS kernel, that one calls
	MUX-14 with the handshaking protocoll, meaning: MUX-14-00. CX=ID,
	BX=VERSION.

	If the magic number was found:
		+ LOAD_UNF assumes that we are running on a FreeDOS kernel and
		verifies the requested NLSFUNC version.
		+ the call passes a pointer to the central nlsInfoBlock, which
		is saved for future use.
		+ The call returns with error.

	Otherwise, the call is passed on unchanged.

	===

	Result: the exit status is ignored, but if the address has been updated,
	it is used to work the chain to search for the requested NLS pkg
	etc. pp.
*/
static int alreadyLoaded(UWORD cp, UWORD cntry)
{	struct nlsPackage far* pkg;

	assert(nlsRoot);

	if((pkg = nlsRoot->chain) != NULL) do {
		if((cntry == NLS_DEFAULT || cntry == pkg->cntry)
		 && (cp == NLS_DEFAULT || cp == pkg->cp))
			return 1;
	} while((pkg = pkg->nxt) != NULL);

	return 0;
}

#ifdef DEBUG
void waitKey(void)
{
	fputs("Press ENTER to proceed...", stdout);
	getchar();
}
#else
#define waitKey()
#endif

#define addtab(idx)	copyTab(d->data[tab[(idx)]], &(pkg->nlsPointers[fctid++]), &data, dataStart)
static void copyTab(struct UNF_DATA_ENTRY *d
 , struct nlsPointer *poi
 , BYTE **data
 , BYTE *dataStart)
{	BYTE *dataEnd, *p;
	struct nlsPackage far* pkg;
	int i;

	assert(nlsRoot);

	poi->subfct = d->fct;
	if((nlsRoot->flags & NLS_CODE_MODIFYABLE_DATA) == 0) {
		/* we can share the same data space */
		/* first search within the currently loaded stuff.
			Use brute force for simplicity. */
		for(p = dataStart, dataEnd = *data - d->size; p <= dataEnd; ++p)
			if(memcmp(p, d->data, d->size) == 0) {
				/* OK found */
				poi->pointer = MK_FP(0, p - dataStart);
				return;
			}

		/* second, try the alrady loaded tables in the nlsChain */
		/* Possible hazard:
			The NLS pkg do not maintain the size of the loaded data,
			that means one only knows where the loaded data starts,
			but not where it ends; therefore one has no absolute
			prove that the found memory area belongs to the
			NLS pkg totally.
			However because most tables are prefixed by a (WORD)
			size and if this word matches the compare, both tables
			do have the same size.
		*/
		if((pkg = nlsRoot->chain) != NULL) do {
			for(i = 0; i < pkg->numSubfct; ++i)
				if(_fmemcmp(pkg->nlsPointers[i].pointer
				 , (BYTE far*)d->data, d->size) == 0) {
				 	poi->pointer = pkg->nlsPointers[i].pointer;
				 	return;
				}
		} while((pkg = pkg->nxt) != NULL);
	}

	/* Otherwise, load the new data bytes locally */
	poi->pointer = MK_FP(0, *data - dataStart);
	memcpy(*data, d->data, d->size);
	*data += d->size;
}

static UWORD DOSAlloc(UWORD size, UWORD mode)
{	int UMBUnlink, oldAllocMode;
	int i;

	_AX = 0x5800;		/* Get Allocation strategy */
	geninterrupt(0x21);
	oldAllocMode = _AX;
	if(_Carry) {
		fputs("Failed to access memory -- Probable memory corruption!\n"
		 , stderr);
		exit(91);
	}

	UMBUnlink = 0;
	if((mode & (0x40 | 0x80)) != 0) {
		/* UMBs are required */
		_AX = 0x5802;	/* Get UMB Link state */
		geninterrupt(0x21);
		i = _AL;
		if(_Carry) {
			/* No UMBs */
			mode &= ~(0x40 | 0x80);
		} else if(i == 0) {
			/* Link in the UMBs */
			UMBUnlink = 1;
			_BX = 1;
			_AX = 0x5803;
			geninterrupt(0x21);
			if(_Carry) {
				/* Something failed --> No UMBs */
				mode &= ~(0x40 | 0x80);
			}
		}
	}

	_BX = mode & 0xFF;
	_AX = 0x5801;
	geninterrupt(0x21);
	if(_Carry) {
		i = 0;
		goto errRet;
	}

	_BX = size;
	_AH = 0x48;
	geninterrupt(0x21);
	i = _AX;
	if(_Carry)		/* Memory allocation failed */
		i = 0;

errRet:
	if(UMBUnlink) {
		_BX = 0;
		_AX = 0x5803;
		geninterrupt(0x21);		/* Ignore rv */
	}
	_BX = oldAllocMode;
	_AX = 0x5801;
	geninterrupt(0x21);

	return i;
}

static int loadFile(char *fnam, WORD cntry, WORD cp)
{	struct UNF_DATA *d;
	int rc, *tab, i, j, maxfct;
	int fctid;
	unsigned pkgSize;
	int subFcts;
	BYTE *dataStart, *data;		/* start of & current highwater of data
									within tail of *pkg */
	UWORD pkgSegm;				/* segment where the NLS pkg is loaded to */

	if((d = unf_load(fnam, cntry, cp, (UNF_Error_Fct* )unf_error)) == NULL) {
		if(!fnam)
			fprintf(stderr, "Failed to load UNF file for %04u:%04u\n"
			 , cntry, cp);
		else if(strcmp(fnam, "-"))
			fprintf(stderr, "Failed to load UNF file from stdin\n");
		else
			fprintf(stderr, "Failed to load UNF file: %s\n", fnam);
		waitKey();
		return 33;
	}

	rc = 0;

	if(unf_kernelCheck(d, (UNF_Error_Fct *)unf_error)) {
		rc = 34;
		goto errRet;
	}

	if(alreadyLoaded(d->codepage, d->country)) {	
		fprintf(stderr, "The NLS pkg %04u:%04u is already loaded.\n"
		 , d->country, d->codepage);
		rc = 35;
		goto errRet;
	}

		/* The kernelCheck() ensures that the standard tables are
			available: 1, 2, 4, 5, 6, 7, 0x23 */
	tab = unf_tab(d);

	/* Calculate the dimension of the tables:
		subFcts := number of available subfunctions
		pkgSize := maximal size of NLS pkg
		maxfct := highest sunfct number
	*/
	i = 256;
	subFcts = pkgSize = maxfct = 0;
	while(i--) if(i != 0x23 && tab[i] != -1) {
		pkgSize += d->data[tab[i]]->size;
		++subFcts;
		if(!maxfct)
			maxfct = i;
	}

	assert(pkgSize);

	if((pkg = malloc(sizeof(struct nlsPackage)	/* basic structure */
	 + sizeof(struct nlsPointer)			/* to be returned except 0x01 */
	   * (subFcts - 2)						/* 1 part of basic; 1 == 0x01 */
	 + 1									/* subfct ID of 0x01 */
	 + pkgSize)) == NULL) {					/* the data of all subfcts */
	 	rc = 40;
	 	fputs("Out of local memory\n", stderr);
	 	goto errRet1;
	}

		/* Where to load the data of subfct #1 to */
	data = (BYTE*)pkg + sizeof(struct nlsPackage)
	 + sizeof(struct nlsPointer) * (subFcts - 2);

	/* Prepare the nlsPackage */
	pkg->cntry	= d->country;
	pkg->cp		= d->codepage;
		/* Because this is the hardcoded package, all data is internally
			retreiveble without the need to channel any request through
			MUX */
	pkg->flags	= NLS_FLAG_HARDCODED;
	pkg->yeschar	= d->data[tab[0x23]]->data[0];
	pkg->nochar 	= d->data[tab[0x23]]->data[1];
		/* Count the number of tables */
	tab[0x23] = -1;		/* table 0x23 is a pseudo-table */
	pkg->numSubfct = subFcts;

	*data++ = 1;		/* subfct ID */
	memcpy(data, d->data[tab[1]]->data, d->data[tab[1]]->size);
	dataStart = (data += d->data[tab[1]]->size);

	fctid = 0;
		/* Now add the standard tables in the standard order */
	addtab(2);
	addtab(4);
	addtab(5);
	if(tab[3] != -1)
		addtab(3);
	for(i = 5; ++i <= maxfct;) if(tab[i] != -1)
		addtab(i);

	/* Now all tables have been loaded into 'pkg'; *data marks the
		end of the required memory block */
	pkgSegm = DOSAlloc(((data - (BYTE*)pkg) + 0xf) / 16, 0x82);
	if(!pkgSegm) {
		fputs("Out of DOS (far) memory\n", stderr);
		rc = 41;
	} else {
#ifdef DEBUG
		printf(stderr, "LOAD_UNF: NLS pkg %03u:%03u loaded at segment %04x\n"
			, pkg->cntry, pkg->cp, pkgSegm);
#endif
		/* First resolve the local pointers */
		for(i = 0; i < pkg->numSubfct - 1; ++i)
			if(FP_SEG(pkg->nlsPointers[i].pointer) == 0)
				pkg->nlsPointers[i].pointer = MK_FP(pkgSegm
				 , FP_OFF(pkg->nlsPointers[i].pointer)
				  + (dataStart - (BYTE*)pkg));

		/* Prepare to chain the NLS pkg into the NLS chain */
		pkg->nxt = nlsRoot->chain;

		/* Now move the NLS pkg there */
		_fmemcpy(MK_FP(pkgSegm, 0), (BYTE far*)pkg, data - (BYTE*)pkg);

		/* Chain it into the NLS chain */
		nlsRoot->chain = MK_FP(pkgSegm, 0);
		if(activate)
			nlsRoot->actPkg = MK_FP(pkgSegm, 0);

		/* Mark the memory block as part of the system */
		*(UWORD far*)MK_FP(pkgSegm - 1, 1) = 8;
		_fmemcpy(MK_FP(pkgSegm - 1, 8), (BYTE far*)"SC NLS P", 8);

		printf("NLS pkg %03u:%03u successfully loaded at segment %04x\n"
		 , pkg->cntry, pkg->cp, pkgSegm);
	}
errRet1:
	unf_free(tab);
errRet:
	unf_freeData(d);
	waitKey();
	return rc;
}


#if 1
static void
usage(const char *progname)
{
    printf("Usage: %s [-a] file.UNF");
    printf("Loads a NLS pkg into memory for the FreeDOS kernel version 20%02d\n.", REVISION_SEQ);
    printf("\n");
    printf("Options:");
    printf("    -a Activate NLS info.");
}
#endif


main(int argc, char **argv)
{	int rv;

#ifdef DEBUG
	puts("LOAD_UNF started");
#endif
	if(!fetchNlsRoot()) {
		waitKey();
		return 90;
	}

#if 1
	if(argc > 1 &&
	   (argv[1][0] == '-' || argv[1][0] == '/') &&
	   (argv[1][1] == 'A' || argv[1][1] == 'a')) {
#else
	if(argc > 1 && stricmp(argv[1], "/a") == 0) {
#endif
		activate = 1;
		--argc;
		++argv;
	} else
		activate = 0;
	switch(argc) {
	case 1:		/* no argument */
		rv =  loadFile("-", UNF_DEFAULT, UNF_DEFAULT);
		break;

	case 2:		/* one argument */
		rv =  loadFile(argv[1], UNF_DEFAULT, UNF_DEFAULT);
		break;

	case 3:		/* two arguments */
		rv =  loadFile(NULL, atoi(argv[1]), atoi(argv[2]));
		break;

	default:
#if 1
		usage("load_unf");
#else
		puts("Useage: LOAD_UNF [/a] file.UNF");
#endif
		return 126;
	}
	return rv;
}
