/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF_DUMP.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	unf_dump()

	Dump the UNF_DATA structure into an UNF file.

	Works like GRAB_UNF basically.

*/

#include "unf.loc"

#ifdef _MICROC_
#endif

#include "unf_post.loc"

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF_DUMP.C 1.3 2000/06/17 22:32:10 ska Exp ska $";
#endif

#if 1
void unf_dump(struct UNF_DATA const * const d, char const *path)
{
    unf_dump0(0, d, path);
}

void unf_dump0(int ispathname, struct UNF_DATA const * const d, char const *path)
{	struct UNF_DATA_ENTRY **h, *q;
    int num;

    if(!d || !d->numData)
        unf__eerror("unf_dump: No data");

    switch(ispathname ? unf_creat2(path, d->country, d->codepage) : unf_create(path, d->country, d->codepage)) {
#else
void unf_dump(struct UNF_DATA const * const d, char const *path)
{	struct UNF_DATA_ENTRY **h, *q;
	int num;

	if(!d || !d->numData)
		unf__eerror("unf_dump: No data");

	switch(unf_create(path, d->country, d->codepage)) {
#endif
	case 1:
		unf__eerror("Cannot create NLS information file for: %u:%u"
		 , d->country, d->codepage);
	case 2:
		unf__eerror("Out of memory");
	default:
		unf__eerror("Unknown error while creating UNF file");
	case 0:			/* OK */
		break;
	}

	/* Dump preample */
	oradix(10);		/* Radix used to created output */
	oend();			/* Global section ends */

	num = d->numData;
	h = (struct UNF_DATA_ENTRY **)&d->data[0];

	do {
		if((q = *h) != NULL) switch(q->fct) {
		case 1:
				/* The data does not contain the subfct byte, whereas the
					structure does */
			oExtInf((struct nlsExtCntryInfo*)&q->data[-1]);
			break;

		case 2:
			oCharTab(2, "normal character uppercase table", (UWORD*)q->data);
			break;

		case 3:
			oCharTab(3, "normal character lowercase table", (UWORD*)q->data);
			break;

		case 4:
			oCharTab(4, "filename character uppercase table", (UWORD*)q->data);
			break;

		case 5:
			oFTerm((struct nlsFnamTerm *)q->data);
			break;

		case 6:
			oCharTab(6, "collating sequence table", (UWORD*)q->data);
			break;

		case 7:
			oDBCS((UWORD*)q->data);		/* Dump table 7 */
			break;

		case 0x23:
			oYesNo(q->data);	/* Dump table 0x23 */
			break;

		default:				/* all other tables are dumped as they are */
			olabelc(q->fct, NULL);
			ott(q->data, q->size);	/* table without special options */
			oend();
			break;
		}
		++h;
	} while(--num);

	unf_close();
}
