/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF_ERR.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	unf_error() and unf_dummy_error()
	--> used by the unf_load() function to report an error back to the
	calling function.

	"unf_error()" dumps a reasonable error message onto screen and lets
	unf_load() return a NULL value.

	"unf_dummy_error()" does nothing, but lets unf_load() return NULL.
	This is the default behaviour, if NULL is passed as error callback
	function to unf_load().

*/

#include "unf.loc"

#ifdef _MICROC_
#endif

#include "unf_post.loc"

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF_ERR.C 1.2 2000/04/29 04:44:46 ska Exp ska $";
#endif

void unf_error(struct UNF_IContext * const ictxt, UNF_Error_ID err
	, char const * const fmt, ...)
{	va_list ap;
	struct UNF_IFile *p;
	int first;

	if(ictxt && ictxt->error_id)	return;	/* ignore secondary errors */

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	putc('\n', stderr);

	if(!ictxt)	return;

	first = 1;
	if((p = ictxt->fin) != NULL) do {
		if(first) {
			fputs("|>\tin", stderr);
			first = 0;
		} else	fputs("|>\tincluded from", stderr);
		fprintf(stderr, " file \"%s\", line %lu\n", p->fname, p->linenr);
	} while((p = p->prv) != NULL);

	ictxt->error_id = err;
}

/* Do nothing function to be used instead of NULL */
void unf_dummy_error(struct UNF_IContext * const ictxt, UNF_Error_ID err
	, char const * const fmt, ...)
{	ictxt->error_id = err;
}

